import re
from munch import Munch
from typing import Dict, List, Optional
import os

from brock.log import get_logger
from brock.exception import ConfigError, UsageError
from brock.config.config import Config
from brock.executors import Executor
from brock.executors.host import HostExecutor
from brock.executors.docker import DockerExecutor
from brock.executors.ssh import SshExecutor


class Option:

    def __init__(self, name: str, option_dict: dict):
        self.name = name
        self.argument = option_dict.get('argument', None)
        self.flag = option_dict.get('flag', None)
        self.help = option_dict.get('help', None)
        self.default = option_dict.get('default', None)
        self.choice = option_dict.get('choices', [])
        self.variable = option_dict.get('variable', None)
        self.short_name = option_dict.get('short_name', None)
        self.option = option_dict.get('option', False)
        self.required = option_dict.get('required', False)

        # Renaming option-name to OPTION_NAME
        if self.variable is None:
            self.env_name = name.upper().replace('-', '_')
        else:
            self.env_name = self.variable


class Command:
    '''Handles user defined commands

    This class provides storage for all the config options related to user
    defined commands and provides a way to execute all the defined steps with
    a correct content.
    '''

    def __init__(self, name: str, config: Munch, default_executor: Optional[str]):
        self._log = get_logger()

        self._chdir = config.get('chdir')
        self._depends_on = config.get('depends_on', [])
        self._default_executor = config.get('default_executor', default_executor)

        self.name = name
        self.help = config.get('help', '')
        self.config_options = config.get('options', None)
        self.options = {}

        steps = []
        for step in config.get('steps', []):
            if isinstance(step, Munch):
                steps.append(step.toDict())
            else:
                steps.append(step)
        self._steps = steps

        if self.config_options is not None:
            option_dict = Munch.toDict(self.config_options)
            for option in option_dict:
                self.options[option] = Option(option, option_dict[option])

    def exec(self, project, step_options, **kwargs) -> int:
        for dependency in self._depends_on:
            exit_code = project.exec(dependency)
            if exit_code != 0:
                return exit_code

        self._log.extra_info(f'Executing command {self.name}')

        for step in self._steps:
            exit_code = self._exec_step(project, step, step_options)

            if exit_code != 0:
                return exit_code

        return exit_code

    def _exec_step(self, project, step, step_options) -> int:
        env_options = {}
        if step_options is not None:
            env_options = self._get_options(step_options)
        if type(step) is str:
            res = re.search(r'(?:^@(\w+) )?(.*)', step)
            if res is None:
                raise ConfigError(f'Unknown step format: {step}')
            executor = res.group(1)
            command = res.group(2)
            if not executor:
                executor = self._default_executor
        elif type(step) is dict:
            executor = step.get('executor', self._default_executor)
            shell = step.get('shell', project.get_default_shell(executor))
            if shell is None:
                raise ConfigError('Shell must be specified')
            command = self._get_shell_command(step.get('script'), shell)
        else:
            raise ConfigError(f'Unexpected step type: {type(step)}')
        return project.exec_raw(command, executor, self._chdir, env_options=env_options)

    def _get_options(self, step_options):
        env_options = {}
        for step_opt in step_options.values():
            for opt, vals in self.options.items():
                if '*' == vals.argument:
                    if step_options[opt.replace('-', '_').lower()] == step_opt:
                        if isinstance(step_opt, tuple):
                            separator = ' '
                            new_step_opt = separator.join(word for word in step_opt if word)
                            env_options[vals.env_name] = new_step_opt
                        else:
                            env_options[vals.env_name] = step_opt
                if vals.flag is not None:
                    if vals.flag == step_opt:
                        env_options[vals.env_name] = vals.flag
                elif vals.argument is not None and '*' != vals.argument:
                    try:
                        if step_options[opt.replace('-', '_').lower()] == step_opt:
                            env_options[vals.env_name] = step_opt
                    except KeyError:
                        continue
                elif vals.option:
                    try:
                        if step_options[opt.replace('-', '_').lower()] == step_opt:
                            env_options[vals.env_name] = step_opt
                    except KeyError:
                        continue
        return env_options

    def _get_shell_command(self, script, shell) -> List[str]:
        if shell in ('sh', 'bash', 'zsh', 'powershell'):
            command = [shell, '-c']
            separator = '; '
        elif shell == 'powershell':
            command = [shell, '-Command']
            separator = '; '
        elif shell == 'cmd':
            command = [shell, '/C']
            separator = ' & '
        else:
            raise ConfigError(f'Unsupported shell: {shell}')

        lines = [ln.strip() for ln in script.splitlines() if ln.strip()]
        command.append(separator.join(lines))

        return command


class Project:
    '''Handles processing of the cli commands related to the project configuration.'''
    _default_executor: Optional[str] = None
    _executors: Dict[str, Executor] = {}
    _commands: Dict[str, Command] = {}
    _default_command: Optional[str] = None
    _prev_executor = None

    def __init__(self, config: Config):
        self._log = get_logger()
        self._default_executor = config.executors.get('default')

        for name, cmd in config.commands.items():
            if name == 'default':
                self._default_command = cmd
                continue
            if len(name.split()) != 1:
                raise ConfigError(f'Command must be a single word: {name}')
            self._commands[name] = Command(name, cmd, self._default_executor)

        if self._default_command is None:
            if len(config.commands) == 1:
                self._default_command = next(iter(config.commands))
        self._executors['host'] = HostExecutor(config, 'host')
        for name, executor in config.executors.items():
            if name == 'default':
                continue
            elif executor.type == 'docker':
                self._executors[name] = DockerExecutor(config, name)
            elif executor.type == 'ssh':
                self._executors[name] = SshExecutor(config, name)
            else:
                raise ConfigError(f"Unknown executor type '{executor.type}'")
        if self._default_executor is None:
            if len(config.executors) == 0:
                self._default_executor = 'host'
            elif len(config.executors) == 1:
                self._default_executor = next(iter(config.executors))
            else:
                self._default_executor = None

    def _get_selected_executors(self, executor_name: Optional[str] = None) -> List[Executor]:
        if executor_name is not None:
            if executor_name not in self._executors:
                raise ConfigError(f'Unknown executor {executor_name}')
            return [self._executors[executor_name]]
        return list(self._executors.values())

    def on_exit(self):
        if self._prev_executor:
            self._executors[self._prev_executor].sync_out()

    @property
    def commands(self) -> Dict[str, Command]:
        return self._commands

    @property
    def default_command(self) -> Optional[str]:
        return self._default_command

    @property
    def executors(self) -> Dict[str, Executor]:
        return self._executors

    @property
    def default_executor(self) -> Optional[str]:
        return self._default_executor

    def get_default_shell(self, executor) -> Optional[str]:
        if executor in self._executors:
            return self._executors[executor].default_shell
        else:
            return None

    def status(self):
        for name, executor in self._executors.items():
            if name != 'host':
                print(f'{name}: {executor.status()}')

    def stop(self, executor_name: Optional[str] = None):
        for executor in self._get_selected_executors(executor_name):
            executor.stop()

    def restart(self, executor_name: Optional[str] = None):
        for executor in self._get_selected_executors(executor_name):
            executor.restart()

    def update(self, executor_name: Optional[str] = None):
        for executor in self._get_selected_executors(executor_name):
            executor.update()

    def exec(self, command: Optional[str] = None, env_options: Optional[dict] = None) -> int:
        if command is None:
            if self._default_command is None:
                raise UsageError('No default command defined')
            command = self._default_command
            self._log.info(f'No command specified, using default ({command})')

        if command not in self._commands:
            raise UsageError(f'Unknown command {command}')
        return self._commands[command].exec(self, env_options)

    def exec_raw(
        self,
        command: str,
        executor_name: Optional[str] = None,
        chdir: Optional[str] = None,
        env_options: Optional[dict] = None
    ) -> int:
        if not executor_name:
            if self._default_executor:
                executor_name = self._default_executor
            else:
                raise ConfigError('No default executor is set!')
        if executor_name not in self._executors:
            raise ConfigError(f'Unknown executor {executor_name}')

        if self._prev_executor != executor_name:
            if self._prev_executor:
                self._executors[self._prev_executor].sync_out()
            self._executors[executor_name].sync_in()
            self._prev_executor = executor_name
        return self._executors[executor_name].exec(command=command, chdir=chdir, env_options=env_options)

    def shell(self, executor_name: str) -> int:
        if executor_name not in self._executors:
            raise ConfigError(f'Unknown executor {executor_name}')

        self._executors[executor_name].sync_in()
        self._prev_executor = executor_name
        return self._executors[executor_name].shell()



import os
import sys

if __name__ == '__main__':
    if __package__ == '':
        # To be able to run 'python wheel-0.9.whl/wheel':
        import os.path
        path = os.path.dirname(os.path.dirname(__file__))
        sys.path[0:0] = [path]

    import brock.cli.main

    sys.exit(brock.cli.main.main())



from __future__ import annotations
import pytest
from pathlib import Path
from copy import deepcopy
import yaml
import os
from unittest.mock import patch
import brock
from brock.cli.main import main
from brock.config.config import Config

project_template = {
    'version': '0.0.9',
    'project': 'test',
    'commands': {},
    'executors': {
        'alpine': {
            'type': 'docker',
            'image': 'alpine',
        },
    },
}
command_template = {'default_executor': 'alpine', 'options': {}, 'steps': [{'script': ''}]}


def mock_config(config: str):
    c = Config(configs=[config])
    setattr(c, 'base_dir', str(Path(__file__).parent))
    setattr(c, 'work_dir', str(Path().cwd()))

    common_prefix = os.path.commonprefix([c.work_dir, c.base_dir])
    work_dir_rel = os.path.relpath(c.work_dir, common_prefix).replace('\\', '/')
    setattr(c, 'work_dir_rel', work_dir_rel)
    return c


def create_config(options, script: str):
    command = deepcopy(command_template)
    command['options'].update(options)
    command['steps'][0]['script'] = script
    project = deepcopy(project_template)
    project['commands'].update({'test_cmd': command})
    return mock_config(yaml.dump(project))


def execute_brock(args: list[any], config: str) -> int:
    with patch('brock.cli.main.Config') as brock_config:
        with patch('brock.cli.main.exit') as exit_fn:
            brock_config.return_value = config
            main(args)

            return exit_fn.call_args[0][0]


@pytest.mark.parametrize(
    'option,args,expected', [
        ({
            'verbose': {
                'flag': '--my_flag'
            }
        }, [], ('VERBOSE', '')),
        ({
            'verbose': {
                'flag': '--my_flag'
            }
        }, ['--verbose'], ('VERBOSE', '--my_flag')),
        ({
            'verbose': {
                'flag': '--my_flag'
            }
        }, ['--verbose', '--verbose'], ('VERBOSE', '--my_flag')),
        ({
            'short-verbose': {
                'flag': '--my-flag',
                'short_name': '-v'
            }
        }, ['-v'], ('SHORT_VERBOSE', '--my-flag')),
    ]
)
def test_brock_options_flag(option, args, expected):
    """ Test combination of flags that should execute without errors. """
    config = create_config(
        option, f'printenv\n ret=$([[ "${{{expected[0]}}}" != "{expected[1]}" ]] && echo 1 || echo 0)\n exit $ret'
    )
    ret = execute_brock(['test_cmd', *args], config)
    assert ret == 0, 'Brock failed to execute a command'


@pytest.mark.parametrize(
    'option,args,expected',
    [
        ({
            'arg1': {
                'argument': 1
            }
        }, ['foo'], ('ARG1', 'foo')),
        ({
            'arg1': {
                'argument': 1
            },
            'arg2': {
                'argument': 2
            }
        }, ['foo', 'boo'], ('ARG2', 'boo')),
        ({
            'arg1': {
                'argument': 1
            },
            'arg2': {
                'argument': 4
            }
        }, ['foo', 'boo'], ('ARG2', 'boo')),  # order is important not the continuity
        ({
            'arg1': {
                'argument': 1
            },
            'arg2': {
                'argument': 2
            }
        }, ['foo'], ('ARG2', '')),
        ({
            'arg1': {
                'argument': 1,
                'required': True
            }
        }, ['foo'], ('ARG1', 'foo')),
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect']
            }
        }, ['sel'], ('ARG1', 'sel')),
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect']
            }
        }, ['ect'], ('ARG1', 'ect')),
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect'],
                'required': True
            }
        }, ['ect'], ('ARG1', 'ect')),
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect'],
                'default': 'ect'
            }
        }, [], ('ARG1', 'ect')),
        ({
            'argN': {
                'argument': '*'
            }
        }, ['arg1', 'arg2'], ('ARGN', 'arg1 arg2')),
        ({
            'argN': {
                'argument': '*'
            }
        }, [], ('ARGN', '')),
        ({
            'argN': {
                'argument': '*',
                'default': 'blob'
            }
        }, [], ('ARGN', 'blob')),
    ]
)
def test_brock_argument_valid(option, args, expected):
    """ Test combination of argument that should execute without errors. """
    config = create_config(
        option, f'printenv\n ret=$([[ "${{{expected[0]}}}" != "{expected[1]}" ]] && echo 1 || echo 0)\n exit $ret'
    )
    ret = execute_brock(['test_cmd', *args], config)
    assert ret == 0, 'Brock failed to execute a command'


@pytest.mark.parametrize(
    'option,args,expected',
    [
        ({
            'arg1': {
                'argument': 1,
                'required': True
            }
        }, [], 'ARG1'),  # argument is required
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect']
            }
        }, ['foo'], 'ARG1'),  # value is not from selection
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect']
            }
        }, [], 'ARG1'),  # value is not from selection
        ({
            'arg1': {
                'argument': 1,
                'choices': ['sel', 'ect'],
                'default': 'ect'
            }
        }, ['foo'], 'ARG1'),  # value is not from selection
    ]
)
def test_brock_argument_invalid(option, args, expected):
    """ Test combination of argument that should error. """
    config = create_config(option, f'printenv\n ret=$([[ "${{{expected}}}" != "" ]] && echo 1 || echo 0)\n exit $ret')
    ret = execute_brock(['test_cmd', *args], config)
    assert ret != 0, 'Brock failed to execute a command'


@pytest.mark.parametrize(
    'option,args,expected',
    [
        ({
            'opt1': {
                'help': ''
            }
        }, ['--opt1=foo'], ('OPT1', 'foo')),
        ({
            'opt1': {
                'short_name': '-o'
            }
        }, ['-o=boo'], ('OPT1', 'boo')),
        ({
            'opt1': {
                'short_name': '-o'
            }
        }, ['-o', 'goo'], ('OPT1', 'goo')),
        ({
            'very-slow': {
                'help': ''
            }
        }, ['--very-slow=yes'], ('VERY_SLOW', 'yes')),
        ({
            'very-slow': {
                'help': ''
            }
        }, ['--very-slow', 'no'], ('VERY_SLOW', 'no')),
        ({
            'very-slow': {
                'required': True
            }
        }, ['--very-slow', 'no'], ('VERY_SLOW', 'no')),
        ({
            'very-slow': {
                'required': True,
                'default': 'goo'
            }
        }, [], ('VERY_SLOW', 'goo')),  # ok even if required there is default
        ({
            'very-slow': {
                'choices': ['foo', 'boo']
            }
        }, ['--very-slow', 'foo'], ('VERY_SLOW', 'foo')),
        ({
            'very-slow': {
                'choices': ['foo', 'boo']
            }
        }, ['--very-slow', 'goo'], ('VERY_SLOW', 'goo')),
        ({
            'very-slow': {
                'choices': ['foo', 'boo'],
                'default': 'foo'
            }
        }, [], ('VERY_SLOW', 'foo')),
        ({
            'very-slow': {
                'required': True,
                'choices': ['foo', 'boo'],
                'default': 'foo'
            }
        }, [], ('VERY_SLOW', 'foo')),  # ok even if required there is default
        ({
            'very-slow': {
                'short_name': '-v',
                'choices': ['foo', 'boo'],
                'default': 'foo'
            }
        }, ['-v', 'foo'], ('VERY_SLOW', 'foo')),
    ]
)
def test_brock_option_valid(option, args, expected):
    """ Test combination of option that should execute without error. """
    config = create_config(
        option, f'printenv\n ret=$([[ "${{{expected[0]}}}" != "{expected[1]}" ]] && echo 1 || echo 0)\n exit $ret'
    )
    ret = execute_brock(['test_cmd', *args], config)
    assert ret == 0, 'Brock failed to execute a command'


@pytest.mark.parametrize(
    'option,args,expected',
    [
        ({
            'very-slow': {
                'required': True
            }
        }, [], 'VERY_SLOW'),  # missing required thing
        ({
            'very-slow': {
                'short_name': '-v',
                'required': True
            }
        }, [], 'VERY_SLOW'),  # no choices without default
        ({
            'very-slow': {
                'choices': ['foo', 'boo'],
                'required': True
            }
        }, [], 'VERY_SLOW'),  # no choices without default
        ({
            'very-slow': {
                'choices': ['foo', 'boo']
            }
        }, ['--very-slow', 'goo'], 'VERY_SLOW'),  # invalid choices
        ({
            'very-slow': {
                'choices': ['foo', 'boo']
            }
        }, ['--very-slow'], 'VERY_SLOW'),  # expected value after option
        ({
            'very-slow': {
                'choices': ['foo', 'boo'],
                'default': 'goo'
            }
        }, [], 'VERY_SLOW'),  # invalid default
    ]
)
def test_brock_option_invalid(option, args, expected):
    """ Test combination of option that should error. """
    config = create_config(option, f'printenv\n ret=$([[ "${{{expected}}}" != "" ]] && echo 1 || echo 0)\n exit $ret')
    ret = execute_brock(['test_cmd', *args], config)
    assert ret != 0, 'Brock failed to execute a command'



class BaseBrockException(Exception):
    ERROR_CODE = -1

    def __init__(self, message: str = ''):
        self.message = message


class ConfigError(BaseBrockException):
    ERROR_CODE = 10


class UsageError(BaseBrockException):
    ERROR_CODE = 20


class ExecutorError(BaseBrockException):
    ERROR_CODE = 30



import os
import sys
import subprocess
import docker
import time

from typing import Optional, Union, Sequence, Dict, List, Any, Union
from brock.log import get_logger
from brock.executors import Executor
from brock.config.config import Config
from brock.exception import ConfigError, ExecutorError


class Container:
    '''Docker container abstraction'''

    def __init__(
        self,
        name: str,
        platform: Optional[str] = 'linux',
        image: Optional[str] = None,
        dockerfile: Optional[str] = None,
        env: Dict[str, Any] = {},
        mac_address: str = None,
        ports: Dict[Union[str, int], int] = {},
        devices: List[str] = [],
        volumes: Dict[str, Dict[str, Any]] = {},
        run_endpoint: str = None
    ):
        self.name = name
        self._platform = platform
        self._dockerfile = dockerfile
        self._env = env
        self._mac_address = mac_address
        self._ports = ports
        self._devices = devices
        self._volumes = volumes
        self._run_endpoint = run_endpoint

        self._log = get_logger()

        if self._dockerfile:
            image_parts = [self.name]
        elif image:
            image_parts = image.split(':')
        else:
            raise ExecutorError('Image or dockerfile must be defined')

        if len(image_parts) == 2:
            self._image_name = image_parts[0]
            self._image_tag = image_parts[1]
        elif len(image_parts) == 1:
            self._image_name = image_parts[0]
            self._image_tag = 'latest'
        else:
            raise ExecutorError('Invalid docker image')

    @property
    def _docker_run(self) -> docker.DockerClient:
        try:
            if self._run_endpoint:
                return docker.DockerClient(base_url=self._run_endpoint)
            else:
                return docker.from_env()
        except docker.errors.DockerException:
            raise ExecutorError('Docker engine is not running')

    @property
    def _docker(self) -> docker.DockerClient:
        try:
            return docker.from_env()
        except docker.errors.DockerException:
            raise ExecutorError('Docker engine is not running')

    @property
    def _container(self):
        try:
            return self._docker.containers.get(self.name)
        except docker.errors.NotFound:
            raise ExecutorError('Docker container not found')

    @property
    def _image(self):
        try:
            return self._docker.images.get(f'{self._image_name}:{self._image_tag}')
        except docker.errors.ImageNotFound:
            return None

    @property
    def _isolation(self) -> Optional[str]:
        if self._platform == 'windows':
            if self._image is None:
                raise ExecutorError('Docker is probably switched to incorrect platform')
            image_version = self._image.attrs['OsVersion'].split('.')[:3]
            info = self._docker.info()
            os_version = info['OSVersion'].split('.')[:3]
            if '.'.join(image_version) == '.'.join(os_version):
                return 'process'
            return 'hyperv'
        return None

    def _create_volumes(self) -> None:
        '''Creates named volumes used by the container'''
        try:
            volumes = map(lambda x: x.name, self._docker.volumes.list())
        except docker.errors.APIError as ex:
            raise ExecutorError(f'Failed to list volumes: {ex}')

        for name in self._volumes:
            # volume tied to physical path
            if os.path.exists(name):
                continue

            if name not in volumes:
                self._log.extra_info(f'Creating volume {name}')
                try:
                    self._docker.volumes.create(name)
                except docker.errors.APIError as ex:
                    raise ExecutorError(f'Failed to create volume: {ex}')

    def _build(self):
        self._log.info(f'Building Docker image from {self._dockerfile}')
        dockerdir = os.path.dirname(self._dockerfile)
        try:
            generator = self._docker.api.build(
                path=dockerdir, platform=self._platform, tag=f'{self._image_name}:{self._image_tag}', decode=True
            )
            try:
                for chunk in generator:
                    if 'stream' in chunk:
                        print(chunk['stream'], end='')
            except KeyboardInterrupt:
                self._log.warning('Execution interrupted')

        except (docker.errors.BuildError, docker.errors.APIError) as e:
            raise ExecutorError(f'Unable to build image: {str(e)}')

    def _pull(self):
        self._log.info(f'Pulling image {self._image_name}:{self._image_tag}')
        try:
            generator = self._docker.api.pull(
                self._image_name, self._image_tag, platform=self._platform, stream=True, decode=True
            )
            try:
                for chunk in generator:
                    if 'progress' not in chunk and 'status' in chunk:
                        print(f"{chunk.get('id', '')} -> {chunk['status']}")
            except KeyboardInterrupt:
                self._log.warning('Execution interrupted')

        except docker.errors.APIError as ex:
            raise ExecutorError(f'Failed to pull image: {ex}')

    def is_running(self) -> bool:
        try:
            self._docker.containers.get(self.name)
        except docker.errors.NotFound as ex:
            return False
        return True

    def start(self) -> None:
        if self.is_running():
            return

        if self._image is None:
            self.update()

        self._log.info(f'Starting container {self.name}')
        self._create_volumes()
        try:
            res = self._docker_run.containers.run(
                image=f'{self._image_name}:{self._image_tag}',
                name=self.name,
                auto_remove=True,
                detach=True,
                stdin_open=True,
                environment=self._env,
                mac_address=self._mac_address,
                ports=self._ports,
                platform=self._platform,
                isolation=self._isolation,
                devices=self._devices,
                volumes=self._volumes,
            )
            self._log.debug(res)
        except docker.errors.APIError as ex:
            raise ExecutorError(f'Failed to start container: {ex}')

    def stop(self) -> None:
        if not self.is_running():
            return
        self._log.info(f'Stopping container {self.name}')
        self._container.stop()

    def update(self) -> None:
        if self._dockerfile:
            self._build()
        else:
            self._pull()

        if self.is_running():
            self.stop()

    def exec(self, command: Union[str, Sequence[str]], work_dir: str) -> int:
        if not self.is_running():
            self.start()

        self._log.extra_info(f'Executing command in container {self.name}: {command}')
        self._log.debug(f'Work dir: {work_dir}')

        try:
            exec_id = self._container.client.api.exec_create(
                self._container.id, command, workdir=work_dir, environment=self._env
            )['Id']
            output = self._container.client.api.exec_start(exec_id, stream=True, demux=True)
            try:
                for chunk in output:
                    if chunk[0]:
                        print(chunk[0].decode('utf-8', 'replace'), end='')
                    if chunk[1]:
                        print(chunk[1].decode('utf-8', 'replace'), end='', file=sys.stderr)
            except KeyboardInterrupt:
                self._log.warning('Execution interrupted')

            res = self._container.client.api.exec_inspect(exec_id)
            exit_code = res['ExitCode']

            self._log.debug(f'Exit code: {exit_code}')
            return exit_code
        except docker.errors.APIError as ex:
            raise ExecutorError(f'Failed to execute command: {ex}')

    def shell(self, shell: str, work_dir: str) -> int:
        if not self.is_running():
            self.start()

        command = ['docker', 'exec', '-it', '-w', work_dir, self.name, shell]

        self._log.extra_info(f'Starting shell ({shell}) in container {self.name}')
        self._log.debug(f'Work dir: {work_dir}')
        return subprocess.run(command, env=self._env).returncode


class DockerExecutor(Executor):
    '''Executor for docker based toolchains

    The executor launches the docker container automatically when needed. If the
    image is not found, it's pulled before first run.
    '''
    _HOST_PATH = '/host'
    _RSYNC_PATH = '/rsync_volume'

    def __init__(self, config: Config, name: str):
        '''Initializes Docker executor

        :param config: A whole brock configuration
        :param name: Name of the executor
        '''
        super().__init__(config, name)

        self._platform = self._conf.get('platform', 'linux')
        self._prepare = self._conf.get('prepare', [])

        self.env_vars.update(self._conf.get('env', {}))

        if self._default_shell is None:
            if self._platform == 'windows':
                self._default_shell = 'cmd'
            else:
                self._default_shell = 'sh'

        if self._platform == 'windows':
            self._mount_dir = f'C:{self._HOST_PATH}'
        else:
            self._mount_dir = self._HOST_PATH
        self._work_dir_rel = config.work_dir_rel.replace('\\', '/')
        self._work_dir = os.path.join(self._mount_dir, self._work_dir_rel).replace('\\', '/')

        if 'sync' in self._conf:
            self._sync_options = self._conf.sync.get('options', None)
            self._sync_filter = self._conf.sync.get('filter', [])
            self._sync_include = self._conf.sync.get('include', [])
            self._sync_exclude = self._conf.sync.get('exclude', [])
            self._sync_type = self._conf.sync.type
        else:
            self._sync_type = None
        self._synced_in = False
        run_endpoint = None
        if self._sync_type == 'rsync':
            self._sync_exclude = self._conf.sync.get('exclude', [])

            self._rsync_volume_name = f'{config.project}-rsync-volume-{self._hashed_base_dir}'
            self._rsync_container = None

            self._rsync_container = Container(
                f'brock-{config.project}-rsync-{self._hashed_base_dir}',
                platform='linux',
                image='eeacms/rsync:2.3',
                volumes={
                    self._rsync_volume_name: {
                        'bind': self._RSYNC_PATH,
                        'mode': 'rw'
                    },
                    self._base_dir: {
                        'bind': self._HOST_PATH,
                        'mode': 'rw'
                    }
                }
            )
            volumes = {self._rsync_volume_name: {'bind': self._mount_dir, 'mode': 'rw'}}
        elif self._sync_type == 'mutagen':
            try:
                contexts = docker.context.ContextAPI.contexts()
                context = next(c for c in contexts if c.name == 'desktop-linux-mutagen')
                run_endpoint = context.endpoints['docker']['Host']
            except:
                self._log.warning('Failed to get mutagen context, is Mutagen Extension for Docker Desktop installed?')

            volumes = {self._base_dir: {'bind': self._mount_dir, 'mode': 'rw'}}
        else:
            volumes = {self._base_dir: {'bind': self._mount_dir, 'mode': 'rw'}}

        dockerfile = None
        if 'dockerfile' in self._conf:
            dockerfile = os.path.join(self._base_dir, self._conf['dockerfile'])

        self._container = Container(
            f'brock-{config.project}-{name}-{self._hashed_base_dir}',
            platform=self._platform,
            image=self._conf.get('image'),
            dockerfile=dockerfile,
            env=self.env_vars,
            mac_address=self._conf.get('mac_address', None),
            ports=self._conf.get('ports'),
            devices=self._conf.get('devices', []),
            volumes=volumes,
            run_endpoint=run_endpoint,
        )

    def sync_in(self):
        if self._sync_type is None:
            return

        if self._sync_type == 'rsync':
            if self._rsync_container is None:
                return
            self._log.extra_info(f'Rsyncing data into docker volume')
            self._rsync(self._HOST_PATH, self._RSYNC_PATH)
        elif self._sync_type == 'mutagen':
            pass
        else:
            raise ExecutorError('Unknown sync type')

        self._synced_in = True

    def sync_out(self):
        if not self._synced_in:
            return

        if self._sync_type == 'rsync':
            if self._rsync_container is None:
                return
            self._log.extra_info(f'Rsyncing data out of docker volume')
            self._rsync(self._RSYNC_PATH, self._HOST_PATH)
        elif self._sync_type == 'mutagen':
            pass
        else:
            raise ExecutorError('Unknown sync type')

        self._synced_in = False

    def status(self) -> str:
        res = 'Stopped'
        if self._container.is_running() and (self._rsync_container is None or self._rsync_container.is_running()):
            res = 'Running'
        res += f'\n\t{self._container.name}'
        if self._rsync_container is not None:
            res += f'\n\t{self._rsync_container.name}'
        return res

    def stop(self):
        self._container.stop()
        if 'sync' in self._conf and self._sync_type == 'rsync' and self._rsync_container:
            self._rsync_container.stop()

    def restart(self) -> int:
        self.stop()
        time.sleep(1)
        return self._start()

    def update(self):
        self._container.update()

    def exec(
        self,
        command: Union[str, Sequence[str]],
        chdir: Optional[str] = None,
        env_options: Optional[dict] = None
    ) -> int:
        if env_options is not None:
            self.env_vars.update(env_options)
        if not self._container.is_running():
            self._log.info('Executor not running -> starting')
            exit_code = self._start()
            if exit_code != 0:
                return exit_code
        if not self._synced_in:
            self.sync_in()

        directory = self._work_dir
        if chdir:
            directory = self._mount_dir + '/' + chdir
        return self._container.exec(command, directory)

    def shell(self) -> int:
        if not self._container.is_running():
            self._log.info('Executor not running -> starting')
            self._start()

        if not self._synced_in:
            self.sync_in()

        shell = self.default_shell
        if shell is None:
            raise ExecutorError('Shell is not defined')

        return self._container.shell(shell, self._work_dir)

    def _start(self) -> int:
        if self._container.is_running():
            return 0

        self.sync_in()
        self._container.start()
        for command in self._prepare:
            exit_code = self._container.exec(command, self._mount_dir)
            if exit_code != 0:
                self._log.error('Failed to execute prepare steps')
                self._container.stop()
                return exit_code
        return 0

    def _rsync(self, src: str, dest: str):
        if self._rsync_container is None:
            return 0

        if self._sync_options is not None:
            options = []
            for option in self._sync_options:
                options.append(f'{option}')
        else:
            # use default options
            options = ['-a', '--delete']

        for filter_ in self._sync_filter:
            options.append(f"--filter '{filter_}'")
        for include in self._sync_include:
            options.append(f"--include '{include}'")
        for exclude in self._sync_exclude:
            options.append(f"--exclude '{exclude}'")

        exit_code = self._rsync_container.exec(f"rsync {' '.join(options)} {src}/ {dest}", '/')
        if exit_code != 0:
            raise ExecutorError(f'Failed to rsync data')
        return exit_code



import logging
import os
import sys
import platform

from colorama import Fore, Style
from colorama.initialise import wrap_stream
from logging import StreamHandler


class ColoredStreamHandler(StreamHandler):
    """
    Colored logger output.
    """

    def __init__(self, stream=None):
        StreamHandler.__init__(self, stream)

    COLORS = {
        'DEBUG': Fore.CYAN,
        'EXTRA_INFO': Fore.GREEN,
        'INFO': Fore.GREEN,
        'STDOUT': Fore.RESET,
        'STDERR': Fore.RED,
        'WARNING': Fore.YELLOW + Style.BRIGHT,
        'ERROR': Fore.RED + Style.BRIGHT,
        'CRITICAL': Fore.RED + Style.BRIGHT,
    }

    def emit(self, record):
        message = self.format(record)
        try:
            self.stream.write(self.COLORS[record.levelname] + message + Style.RESET_ALL + '\n')
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


class DefaultFormatter(logging.Formatter):

    def __init__(self, fmt='%(message)s'):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):
        ret = logging.Formatter.format(self, record)

        if record.levelno not in (STDOUT, STDERR):
            ret = f'>> {ret}'

        return ret


def colored_handler_factory():
    """
    Factory method for custom stream handler that supports
    colored output on GitlabRunner and correct platform recognition.
    """

    on_windows = platform.system() == 'Windows'
    on_gitlab_ci = os.environ.get('GITLAB_CI', False)

    if on_windows and not on_gitlab_ci:
        log_stream = wrap_stream(sys.stdout, None, None, None, True)
    else:
        log_stream = sys.stdout

    return ColoredStreamHandler(log_stream)


DEFAULT_LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'base_formatter': {
            '()': 'brock.log.DefaultFormatter',
        },
    },
    'handlers': {
        'color_handler': {
            '()': 'brock.log.colored_handler_factory',
            'level': 'DEBUG',
            'formatter': 'base_formatter',
        },
        'normal_handler': {
            'formatter': 'base_formatter',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
        },
    },
    'loggers': {
        'color': {
            'level': 'DEBUG',
            'handlers': ['color_handler'],
            'propagate': False,
        },
        'normal': {
            'level': 'INFO',
            'handlers': ['normal_handler'],
            'propagate': False,
        },
    },
}

INFO = logging.INFO
CRITICAL = logging.CRITICAL
FATAL = logging.FATAL
ERROR = logging.ERROR
WARNING = logging.WARNING
WARN = logging.WARN
STDERR = logging.INFO + 2
STDOUT = logging.INFO + 1
INFO = logging.INFO
EXTRA_INFO = logging.INFO - 5
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET

logging._levelToName[EXTRA_INFO] = 'EXTRA_INFO'
logging._nameToLevel['EXTRA_INFO'] = EXTRA_INFO

logging._levelToName[STDOUT] = 'STDOUT'
logging._nameToLevel['STDOUT'] = STDOUT

logging._levelToName[STDERR] = 'STDERR'
logging._nameToLevel['STDERR'] = STDERR


class Logger(logging.Logger):

    def extra_info(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'EXTRA_INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.extra_info("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(EXTRA_INFO):
            self._log(EXTRA_INFO, msg, args, **kwargs)

    def stdout(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'STDOUT'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.stdout("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(STDOUT):
            self._log(STDOUT, msg, args, **kwargs)

    def stderr(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'STDERR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.stderr("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(STDERR):
            self._log(STDERR, msg, args, **kwargs)


logging.setLoggerClass(Logger)


def init_logging():
    logging.config.dictConfig(DEFAULT_LOGGING)
    logger = get_logger()
    logger.setLevel(INFO)


def disable_color():
    logger = get_logger()
    logger.removeHandler(logger.handlers[0])
    normal = Logger.manager.getLogger('normal')
    logger.addHandler(normal.handlers[0])


def set_verbosity(verbosity):
    logger = get_logger()
    logger.setLevel(verbosity)


def get_logger():
    return Logger.manager.getLogger('color')



import pytest

from brock import __version__
from brock.config.config import Config
from brock.exception import ConfigError


def test_example_config():
    '''Test the config attached as an example.'''
    config = Config(['example_brock.yml'])
    assert str(config.version) == '0.1.0'
    assert config.project == 'someprojectname'
    assert config.help == 'brock --help message'

    assert len(config.commands.keys()) == 5
    assert config.commands.default == 'build'

    assert config.commands.clean.default_executor == 'atollic'
    assert config.commands.clean.chdir == 'foo/bar'
    assert config.commands.clean.help == 'Help message for this command'
    assert config.commands.clean.steps == ['make clean']

    assert config.commands.build.depends_on == ['clean']
    assert config.commands.build.steps == [
        '@atollic make', 'make tests', '@host echo \'Building finished\'', 'run ${VERY_FAST}', 'run ${VERY_SLOW}',
        'run ${SPEED}'
    ]

    assert config.commands.rebuild.depends_on == ['clean', 'build']

    assert len(config.commands.service.steps) == 1
    assert config.commands.service.steps[0].executor == 'atollic'
    assert config.commands.service.steps[0].shell == 'powershell'
    assert config.commands.service.steps[0].script == \
        '$ServiceName = \'EventLog\'\n$ServiceInfo = Get-Service -Name $ServiceName\nWrite-Output $ServiceInfo\n'

    assert len(config.executors.keys()) == 5

    assert config.executors.default == 'python'

    assert config.executors.atollic.type == 'docker'
    assert config.executors.atollic.help == 'Executor --help message'
    assert config.executors.atollic.dockerfile == 'path/to/Dockerfile'
    assert config.executors.atollic.platform == 'windows'
    assert len(config.executors.atollic.env.keys()) == 2
    assert config.executors.atollic.env.SOME_VAR == 'foo'
    assert config.executors.atollic.env.OTHER_VAR == 123
    assert config.executors.atollic.mac_address == '88:99:aa:bb:cc:dd'
    assert config.executors.atollic.ports == {5000: 5000, '6000/udp': 6000}
    assert config.executors.atollic.devices == ['class/{interface class GUID}']
    assert config.executors.atollic.default_shell == 'powershell'

    assert config.executors.python.type == 'docker'
    assert config.executors.python.image == 'python:3.9'
    assert config.executors.python.sync.type == 'rsync'
    assert config.executors.python.sync.options == ['-avm']
    assert config.executors.python.sync.filter == ['+ foo/', '+ foo/**/', '+ foo/**.c', '+ foo/src/**', '- *']
    assert config.executors.python.sync.include == ['foo/bar']
    assert config.executors.python.sync.exclude == ['foo/bar']
    assert config.executors.python.devices == ['/dev/ttyUSB0:/dev/ttyUSB0:rwm']
    assert config.executors.python.prepare == [
        'pip install -r requirements.txt',
        'echo "Foo bar"',
    ]

    assert config.executors.gcc.type == 'docker'
    assert config.executors.gcc.image == 'gcc'
    assert config.executors.gcc.sync.type == 'mutagen'

    assert config.executors.remote.type == 'ssh'
    assert config.executors.remote.host == 'somesite.example.com:1235'
    assert config.executors.remote.username == 'foobar'
    assert config.executors.remote.password == 'strongpassword'


def test_empty_config():
    '''Test empty config raises error (#5813).'''
    with pytest.raises(ConfigError) as ex:
        config = Config(['\n'])
    assert ex.value.message == 'Invalid config file: Config file is empty'


def test_invalid_version():
    '''Test invalid config version raises error - config version must be lower or equal to brock version.'''
    with pytest.raises(ConfigError) as ex:
        config = Config(['version: 5.0.0\nproject: test\n'])
    assert ex.value.message.startswith('Current config requires Brock of version at least')


def test_multiple_configs_in_one_dir():
    '''Test multiple config files in one directory raises error.'''
    with pytest.raises(ConfigError) as ex:
        config = Config(config_file_names=['example_brock.yml', '.brock.yml'])
    assert ex.value.message.startswith('Multiple brock config files found in')


def test_empty_executors():
    '''Test executors config section is empty or missing.'''
    config = Config(['version: 0.0.6\nproject: test\nexecutors: {}\n'])
    assert len(config.executors) == 0

    config = Config(['version: 0.0.6\nproject: test\n'])
    assert len(config.executors) == 0



import os
from typing import Optional, Union, Sequence
from fabric import Connection
from brock.executors import Executor
from brock.config.config import Config
from brock.exception import ExecutorError


class SshExecutor(Executor):
    '''Executor for launching commands on the remote system over ssh'''

    def __init__(self, config: Config, name: str):
        '''Initializes SSH executor

        :param config: A whole brock configuration
        :param name: Name of the executor
        '''
        super().__init__(config, name)

        self._host = self._conf.host
        self._username = self._conf.get('username')
        self._password = self._conf.get('password')

        if self._default_shell is None:
            self._default_shell = 'sh'

    def exec(
        self,
        command: Union[str, Sequence[str]],
        chdir: Optional[str] = None,
        env_options: Optional[dict] = None
    ) -> int:

        self._log.extra_info(f'Executing command on SSH host {self._host}: {command}')

        if type(command) is not str:
            command = ' '.join([f'"{c}"' if ' ' in c else c for c in command])

        if self._username is None:
            username = input('Enter username: ')
        else:
            username = self._username

        if self._password is None:
            password = input('Enter password: ')
        else:
            password = self._password

        if chdir is not None:
            self._log.debug(f'Work dir: {chdir}')
            cmd = f'cd {chdir}; {command}'
        else:
            cmd = command

        try:
            conn = Connection(
                host=self._host, user=username, connect_kwargs={
                    'password': password,
                }, inline_ssh_env=True
            )

            if env_options is not None:
                result = conn.run(cmd, hide=True, env=env_options)
            else:
                result = conn.run(cmd, hide=True)
            print(result.stdout, end='')

            return result.exited
        except Exception as ex:
            raise ExecutorError(f'Failed to run command: {ex}')



import subprocess
import os
import sys
import platform

from typing import Optional, Union, Sequence
from brock.executors import Executor
from brock.config.config import Config
from brock.exception import ExecutorError


class HostExecutor(Executor):
    '''Executor for local host access'''

    def __init__(self, config: Config, name: str):
        '''Initializes Host executor

        :param config: A whole brock configuration
        :param name: Name of the executor
        '''
        super().__init__(config, name)

        self.help = 'Execute command on host computer'

        if self._default_shell is None:
            if platform.system() == 'Windows':
                self._default_shell = 'cmd'
            else:
                self._default_shell = 'sh'

    def exec(
        self,
        command: Union[str, Sequence[str]],
        chdir: Optional[str] = None,
        env_options: Optional[dict] = None
    ) -> int:
        if env_options is not None:
            os.environ.update(env_options)

        os.environ['PYTHONUNBUFFERED'] = '1'

        self._log.extra_info(f'Executing command on host: {command}')
        if not chdir:
            chdir = '.'

        try:
            proc = subprocess.Popen(
                command,
                cwd=os.path.join(self._base_dir, chdir),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except FileNotFoundError:
            raise ExecutorError(f"Unable to run '{command}', binary not found")

        while proc.poll() is None:
            line = proc.stdout.readline()  # type:ignore
            if line:
                print(line.decode(), end='')

            line = proc.stderr.readline()  # type:ignore
            if line:
                print(line.decode(), end='', file=sys.stderr)

        return proc.returncode



import os
import hiyapyco
from schema import Schema, And, Or, Use, Optional, Regex, SchemaError
import typing as t
from munch import Munch
from pathlib import Path

from brock import __version__
from brock.exception import ConfigError
from brock.log import get_logger


class Config(Munch):
    SCHEMA = {
        'version': And(str, Regex(r'^[0-9]+.[0-9]+.[0-9]+$')),
        'project': str,
        Optional('help'): str,
        Optional('default_cmd'): str,
        Optional('commands', default={}): {
            Optional('default'): str,
            str: {
                Optional('default_executor'): str,
                Optional('chdir'): str,
                Optional('help'): str,
                Optional('depends_on'): [str],
                Optional('options'): {
                    Optional(str):
                        Or({
                            'flag': Use(str),
                            Optional('default'): any,
                            Optional('short_name'): any,
                            Optional('variable'): str,
                            Optional('help'): str
                        }, {
                            'argument': Use(str),
                            Optional('default'): any,
                            Optional('required'): bool,
                            Optional('choices'): [any],
                            Optional('variable'): str,
                            Optional('help'): str
                        }, {
                            Optional('option', default=True): Use(str),
                            Optional('default'): any,
                            Optional('short_name'): any,
                            Optional('choices'): [any],
                            Optional('variable'): str,
                            Optional('required'): bool,
                            Optional('help'): str
                        })
                },
                Optional('steps'): [Or(str, {
                    Optional('executor'): str,
                    Optional('shell'): str,
                    'script': str
                })],
            }
        },
        Optional('executors', default={}): {
            Optional('default'):
                str,
            Optional(str):
                Or({
                    'type':
                        'docker',
                    Optional('help'):
                        str,
                    Or('image', 'dockerfile', only_one=True):
                        str,
                    Optional('platform'):
                        str,
                    Optional('env'): {
                        str: Or(str, int, float)
                    },
                    Optional('mac_address'):
                        And(str, Regex(r'^([0-9A-Fa-f]{2}:){5}([0-9A-Fa-f]{2})$')),
                    Optional('ports', default={}): {
                        Or(int, And(str, Regex(r'^\d+/(tcp|udp|sctp)'))): int
                    },
                    Optional('devices'): [str],
                    Optional('sync'):
                        Or(
                            {
                                'type': 'rsync',
                                Optional('options'): [str],
                                Optional('filter'): [str],
                                Optional('include'): [str],
                                Optional('exclude'): [str],
                            },
                            {'type': 'mutagen'},
                        ),
                    Optional('prepare'): [str],
                    Optional('default_shell'):
                        str,
                }, {
                    'type': 'ssh',
                    Optional('help'): str,
                    'host': str,
                    Optional('username'): str,
                    Optional('password'): Use(str)
                })
        }
    }

    def __init__(self, configs: t.Optional[t.List[str]] = None, config_file_names: t.Optional[t.List[str]] = None):
        self._log = get_logger()

        if configs is None:
            configs = self._scan_files(config_file_names)

        merged_config = self._load(configs)
        validated_config = self._validate(merged_config)

        self.update(Munch.fromDict(validated_config))

    def _load(self, configs: t.List[str]) -> Munch:
        try:
            self._log.extra_info('Merging config files')
            config = hiyapyco.load(configs, method=hiyapyco.METHOD_MERGE)
            self._log.debug(f'Merged config: {config}')
        except Exception as ex:
            raise ConfigError(f'Failed to process config files: {ex}')

        if config is None:
            raise ConfigError('Invalid config file: Config file is empty')

        return config

    def _validate(self, config: t.Dict) -> t.Dict:
        try:
            config_schema = Schema(self.SCHEMA)
            config = config_schema.validate(config)
        except SchemaError as ex:
            raise ConfigError(f'Invalid config file: {ex}')

        current_ver = __version__.split('.')
        config_ver = str(config['version']).split('.')
        for pos, val in enumerate(current_ver[0:3]):
            if int(config_ver[pos]) < int(val):
                break
            if int(config_ver[pos]) > int(val):
                raise ConfigError(
                    f'Current config requires Brock of version at least {config["version"]}, you are using {__version__}'
                )

        return config

    def _scan_files(self, file_names: t.Optional[t.List[str]] = None) -> t.List[str]:
        if file_names is None:
            file_names = ['.brock.yml', 'brock.yml', '.brock.yaml', 'brock.yaml']

        self.work_dir = os.getcwd().replace('\\', '/')

        self._log.extra_info(f'Scanning config files, work dir: {self.work_dir}')

        config_files = []
        path_parts = self._split_path(self.work_dir)

        for i in range(1, len(path_parts) + 1):
            found = 0
            for config_file_name in file_names:
                path = Path('').joinpath(*path_parts[:i], config_file_name)

                if path.is_file():
                    self._log.debug(f'Found config file: {path.as_posix()}')
                    config_files.append(path.as_posix())
                    found += 1

            if found > 1:
                raise ConfigError(
                    f"Multiple brock config files found in '{Path('').joinpath(*path_parts[:i]).as_posix()}'"
                )

        if not config_files:
            raise ConfigError(
                f"No config file ({', '.join(file_names)}) found in '{self.work_dir}' or parent directories"
            )

        self.base_dir = os.path.dirname(config_files[0])
        self._log.debug(f'Base dir: {self.base_dir}')

        common_prefix = os.path.commonprefix([self.work_dir, self.base_dir])
        self.work_dir_rel = os.path.relpath(self.work_dir, common_prefix).replace('\\', '/')
        self._log.debug(f'Relative work dir: {self.work_dir_rel}')

        return config_files

    @classmethod
    def _split_path(cls, path):
        all_parts = []
        while True:
            parts = os.path.split(path)
            if parts[0] == path:  # sentinel for absolute paths
                all_parts.insert(0, parts[0])
                break
            elif parts[1] == path:  # sentinel for relative paths
                all_parts.insert(0, parts[1])
                break
            else:
                path = parts[0]
                all_parts.insert(0, parts[1])

        return all_parts



import hashlib
from typing import Optional, Union, Sequence, Dict, Any
from brock.log import get_logger
from brock.config.config import Config
from brock.exception import ExecutorError


class Executor:
    '''Abstract class for all executors

    Provides the common functions all executors must implement or use default
    implementation defined here.
    '''

    def __init__(self, config: Config, name: str):
        self._log = get_logger()

        self._base_dir = config.base_dir
        self._hashed_base_dir = hashlib.md5(self._base_dir.encode('ascii')).hexdigest()

        self._conf = config.executors.get(name)
        self._default_shell = None

        self._env_vars = {
            'BROCK_HOST_PATH': config.work_dir,
            'BROCK_HOST_PROJECT_PATH': config.base_dir,
            'BROCK_RELATIVE_PATH': config.work_dir_rel,
        }

        self.help = ''

        if self._conf is not None:
            self._default_shell = self._conf.get('default_shell')
            self.help = self._conf.get('help', '')

    @property
    def default_shell(self) -> Optional[str]:
        return self._default_shell

    @property
    def env_vars(self) -> Dict:
        return self._env_vars

    def sync_in(self):
        '''Synchronizes local data to executor if needed'''
        pass

    def sync_out(self):
        '''Synchronizes local data out of executor if needed'''
        pass

    def status(self) -> str:
        return 'Idle'

    def stop(self):
        pass

    def restart(self):
        return 0

    def update(self):
        pass

    def exec(
        self,
        command: Union[str, Sequence[str]],
        chdir: Optional[str] = None,
        env_options: Optional[dict] = None
    ) -> int:
        raise NotImplementedError

    def shell(self) -> int:
        '''Opens a shell session, if available'''
        raise ExecutorError("This executor doesn't support direct shell access")



import click

import brock.log as log
from brock import __version__
from brock.cli.analytics import init_analytics


class State:

    def __init__(self, project):
        self.verbosity = 0
        self.no_color = False
        self.project = project
        self.error = None


def set_verbosity(ctx, param, value):
    if value == 0:
        log.set_verbosity(log.INFO)
    elif value == 1:
        log.set_verbosity(log.EXTRA_INFO)
    else:
        log.set_verbosity(log.DEBUG)

    state = ctx.find_object(State)
    if state:
        state.verbosity = value

    return value


def set_analytics(ctx, param, analytics_disabled):
    if not analytics_disabled:
        init_analytics(False)


def set_analytics_dev(ctx, param, enable_dev_analytics):
    if enable_dev_analytics:
        init_analytics(True)


def set_no_color(ctx, param, value):
    if value:
        log.disable_color()

    state = ctx.find_object(State)
    if state:
        state.no_color = value

    return value


pass_state = click.make_pass_decorator(State)



import logging

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration

import brock

trace = sentry_sdk.trace
breadcrumb = sentry_sdk.add_breadcrumb


def trace_entry(func):

    def wrapper(*args, **kwargs):
        with sentry_sdk.start_transaction(op=func.__name__, name=func.__name__):
            return func(*args, **kwargs)

    return wrapper


def init_analytics(enable_dev_analytics: bool) -> None:
    env = 'devel' if enable_dev_analytics else 'main'
    release = None if enable_dev_analytics else brock.__version__
    sentry_sdk.init(
        # This is a public information
        dsn='https://75fe6c1dbd9ffcffebc2a9df1bf793cd@sentry.benderrobotics.com/4',
        environment=env,
        release=release,
        include_local_variables=True,
        before_send=_before_send,
        debug=False,
        integrations=[
            LoggingIntegration(
                level=logging.INFO,  # Capture info and above as breadcrumbs
                event_level=logging.CRITICAL  # Send records as events
            )
        ]
    )


def _before_send(event, hint):
    if 'exc_info' in hint:
        _, exc_value, tb = hint['exc_info']
        if isinstance(exc_value, KeyboardInterrupt):
            return None

    return event



import click
from typing import Optional

from brock.exception import UsageError, BaseBrockException
from brock.log import get_logger, init_logging
from .state import State, pass_state
from click.exceptions import ClickException

init_logging()
log = get_logger()


class ArgumentWithHelp(click.Argument):

    def __init__(self, param_decls, required=None, help=None, **attrs):
        super().__init__(param_decls, required=required, **attrs)
        self.help = help

    def get_help_record(self, ctx):
        """Return help record for custom argument if needed by a custom help command."""
        if self.help:
            return (f'<{self.human_readable_name}>', self.help)
        return None


def create_command(cmd: str, help: Optional[str] = None):
    '''Returns click command function for brock commands'''

    @click.command(name=cmd, help=help)
    @pass_state
    def f(state):
        return state.project.exec(cmd)

    return f


def create_command_with_options(cmd: str, options: dict, help: Optional[str] = None):
    '''Returns click command function for brock commands'''
    click_options = []
    for option_n, value in options.items():
        if value.flag is not None:
            if value.short_name is not None:
                click_options.append(
                    click.Option(
                        param_decls=[f'--{option_n}', value.short_name], flag_value=value.flag, help=value.help
                    )
                )
            else:
                click_options.append(
                    click.Option(param_decls=[f'--{option_n}'], flag_value=value.flag, help=value.help)
                )
        elif value.argument is not None:
            if '*' == value.argument:
                click_options.append(
                    ArgumentWithHelp(param_decls=[option_n], required=value.required, nargs=-1, help=value.help)
                )
            else:
                click_options.insert(
                    int(value.argument) - 1,
                    ArgumentWithHelp(param_decls=[option_n], required=value.required, help=value.help)
                )
        elif value.option:
            if value.choice:
                mapping_table = str.maketrans({',': '|', ' ': '', "'": '', '[': '(', ']': ')'})
                help_opt = f'{value.help} {str(value.choice).translate(mapping_table)}'
            else:
                help_opt = f'{value.help}'
            if value.required and value.default is not None:
                if value.short_name is not None:
                    click_options.append(
                        click.Option(
                            param_decls=[f'--{option_n}', value.short_name], help=help_opt, flag_value=value.default
                        )
                    )
                else:
                    click_options.append(
                        click.Option(param_decls=[f'--{option_n}'], help=help_opt, default=value.default)
                    )
            else:
                if value.short_name is not None:
                    click_options.append(
                        click.Option(
                            param_decls=[f'--{option_n}', value.short_name], help=help_opt, required=value.required
                        )
                    )
                else:
                    click_options.append(
                        click.Option(param_decls=[f'--{option_n}'], help=help_opt, required=value.required)
                    )

    @click.command(name=cmd, help=help, params=click_options)
    @pass_state
    def f(state, **kwargs):
        for option_n, value in options.items():
            current_value_name = value.name.replace('-', '_').lower()
            if current_value_name in kwargs:
                if value.option and value.short_name is not None:
                    try:
                        if '=' in kwargs[current_value_name]:
                            kwargs[current_value_name] = kwargs[value.name].replace('=', '')
                    except TypeError:
                        pass
                if value.default is not None and (
                    kwargs[current_value_name] is None or 0 == len(kwargs[current_value_name])
                ):
                    kwargs[current_value_name] = value.default
                if value.argument is not None:
                    if value.choice or (kwargs[value.name.lower()] is None and value.choice):
                        val = check_choice(kwargs[value.name.lower()], value.choice)

        step_options = {key: kwargs[key] for key in kwargs if kwargs[key] is not None}
        return state.project.exec(cmd, step_options)

    return f


def check_choice(option, choice_list):
    if choice_list:
        if option not in choice_list:
            raise BaseBrockException('Provided option is not in a set choice list')
        else:
            return True


@click.command()
@click.argument('executor', required=False)
@pass_state
def shell(state: State, executor=None):
    '''Open shell in executor'''
    if not executor:
        executor = state.project.default_executor
        if not executor:
            raise UsageError('Multiple executors available, you have to specify which one to use')
    elif executor[0] != '@':
        raise UsageError('Unknown executor name format, use @name')
    else:
        executor = executor[1:]

    return state.project.shell(executor)


@click.command()
@click.argument('input', nargs=-1, type=click.Path())
@pass_state
def exec(state: State, input=None):
    '''Run command in executor'''
    if len(input) == 0:
        raise UsageError('No command specified')
    elif input[0][0] != '@':
        executor = state.project.default_executor
        if not executor:
            raise UsageError('Multiple executors available, you have to specify which one to use')
        executor = '@' + executor
        command = input
    else:
        if len(input) < 2:
            raise UsageError('No command specified')
        executor = input[0]
        command = input[1:]

    executor = executor[1:]
    return state.project.exec_raw(' '.join(command), executor)



import sys
import click
import collections
import logging.config
from typing import Dict, Tuple
from click.exceptions import ClickException
from .state import State, set_verbosity, set_no_color, set_analytics, set_analytics_dev

from brock.exception import BaseBrockException, ConfigError, UsageError
from brock.project import Project
from brock.config.config import Config
from brock import __version__
from brock.log import get_logger, init_logging
from .commands import create_command, shell, exec, create_command_with_options


class CustomCommandGroup(click.Group):
    '''Custom click group for customized help formatting

    The epilog is replaces with custom_epilog with commands like formatting
    with multiple sections possible (e.g. {'Executors': [('python', 'help msg')]})

    The commands are listed in order in they were added
    '''
    custom_epilog: Dict[str, Tuple[str, str]] = {}

    def __init__(self, name=None, commands=None, **attrs):
        super(CustomCommandGroup, self).__init__(name, commands, **attrs)
        #: the registered subcommands by their exported names.
        self.commands = commands or collections.OrderedDict()

    def format_epilog(self, ctx, formatter):
        for section, content in self.custom_epilog.items():
            with formatter.section(section):
                formatter.write_dl(content)

    def list_commands(self, ctx):
        return self.commands


def analytics_options_decorator(func):
    if 'dev' not in __version__:
        return click.option(
            '--disable-analytics',
            is_flag=True,
            expose_value=False,
            default=False,
            callback=set_analytics,
            help='Disable analytics features.'
        )(func)
    else:
        return click.option(
            '--enable-dev-analytics',
            is_flag=True,
            expose_value=False,
            default=False,
            callback=set_analytics_dev,
            help='Force enable analytics during development.'
        )(func)


@click.group(cls=CustomCommandGroup, invoke_without_command=True)
@click.version_option(__version__)
@click.option('--stop', is_flag=False, flag_value='all', default=None, help='Stop project', metavar='[EXECUTOR]')
@click.option(
    '--update',
    is_flag=False,
    flag_value='all',
    default=None,
    help='Update the executor (pull docker image, ...)',
    metavar='[EXECUTOR]'
)
@click.option(
    '-r',
    '--restart',
    is_flag=False,
    flag_value='all',
    default=None,
    help='Restart project (e.g. to reload config)',
    metavar='[EXECUTOR]'
)
@click.option('-s', '--status', is_flag=True, help='Show state of the project')
@click.option('-v', '--verbose', count=True, help='Set logging verbosity', expose_value=False, callback=set_verbosity)
@click.option(
    '--no-color',
    is_flag=True,
    help='Disable default color output',
    multiple=False,
    expose_value=False,
    callback=set_no_color
)
@analytics_options_decorator
@click.pass_context
def cli(ctx, stop, update, restart, status, **kwargs):
    state = ctx.find_object(State)
    # allow running --help and --version even if config parsing failed
    if state.error:
        raise state.error

    if ctx.invoked_subcommand is None:
        if stop:
            ctx.obj.project.stop(None if stop == 'all' else stop)
        elif update:
            ctx.obj.project.update(None if update == 'all' else update)
        elif restart:
            ctx.obj.project.restart(None if restart == 'all' else restart)
        elif status:
            ctx.obj.project.status()
        else:
            # default command if available
            state.project.exec()
    elif stop or update or restart or status:
        raise UsageError('Invalid arguments combination')


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    init_logging()
    log = get_logger()
    project = None
    exit_code = 0

    try:
        config_error = None
        try:
            config = Config()
            project = Project(config)
        except ConfigError as e:
            config_error = e

        state = State(project)
        state.error = config_error

        cli.add_command(shell)
        cli.add_command(exec)

        if project:
            for name, cmd in project.commands.items():
                help = cmd.help + (' (default)' if name == project.default_command else '')
                if cmd.options:
                    cli.add_command(create_command_with_options(name, cmd.options, help.strip()))
                else:
                    cli.add_command(create_command(name, help.strip()))
            cli.help = config.get('help', '')

            executors = []
            for name, executor in project.executors.items():
                help = executor.help + (' (default)' if name == project.default_executor else '')
                executors.append((name, help.strip()))
            cli.custom_epilog = {'Executors': executors}

        cli_error = None
        try:
            result = cli(obj=state, standalone_mode=False, args=args)
        except RuntimeError:
            # * Exit and Abort from click
            result = None
        except Exception as ex:
            cli_error = ex

        if config_error:
            raise config_error
        elif cli_error:
            raise cli_error
        elif result is not None:
            exit_code = result
    except ClickException as ex:
        log.error(ex.message)
        exit_code = ex.exit_code
    except BaseBrockException as ex:
        if len(ex.message) > 0:
            log.error(ex.message)
        exit_code = ex.ERROR_CODE

    if project:
        project.on_exit()
    exit(exit_code)
